import typing
from typing import TYPE_CHECKING

import numpy as np

from test.assertions import QuaternionTest

if TYPE_CHECKING:
    import pyquat.pyquat.wahba as wahba
else:
    import pyquat.wahba as wahba

class TestWahba(QuaternionTest):

    def test_attitude_profile_matrix(self) -> None:
        """attitude_profile_matrix() doesn't raise errors when given observation and reference vectors"""
        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[0.0, 0.0],
                        [1.0, 0.0],
                        [0.0, 1.0]])
        B = wahba.attitude_profile_matrix(obs = obs, ref = ref)
        # Needs actual test here


    def test_davenport_matrix(self) -> None:
        """davenport_matrix() called on the output of attitude_profile_matrix() produces a valid Davenport K matrix"""
        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[0.0, 0.0],
                        [1.0, 0.0],
                        [0.0, 1.0]])
        B = wahba.attitude_profile_matrix(obs = obs, ref = ref)
        K = wahba.davenport_matrix(B)
        self.assertEqual(K.shape[0], 4)
        self.assertEqual(K.shape[1], 4)

        # Test q-method davenport matrix
        K_qekf = wahba.davenport_matrix(B, covariance_analysis = True)
        self.assertEqual(K_qekf[0,0], 0.0)
        np.testing.assert_equal(K_qekf[0,1:3], K[0,1:3])
        for ii in range(0,3):
            self.assertNotEqual(K_qekf[ii,ii], K[ii,ii])



import typing
from typing import TYPE_CHECKING

import numpy as np
from scipy import linalg

from test.assertions import QuaternionTest

if TYPE_CHECKING:
    import pyquat.pyquat as pq
    import pyquat.pyquat.random as random
else:
    import pyquat as pq
    import pyquat.random as random

class TestRandom(QuaternionTest):
    def test_uniform_random_axis(self) -> None:
        """A unit axis is generated by uniform_random_axis()"""
        v = random.uniform_random_axis()
        uv = v / linalg.norm(v)
        np.testing.assert_array_almost_equal(v, uv, decimal=12)

    def test_randu(self) -> None:
        """urand() produces random numbers which are inside the correct range (-1,+1). """
        for ii in range(0,100):
            v = random.randu()
            self.assertLessEqual(-1.0, v)
            self.assertGreaterEqual(1.0, v)

    def test_rand(self) -> None:
        """rand() produces a quaternion on the unit sphere"""
        qr = random.rand()
        self.assertEqual(type(qr), pq.Quat)

        v = qr.to_vector()
        self.assertAlmostEqual(linalg.norm(v), 1.0, places=15)

        for ii in range(0,4):
            self.assertLessEqual(-1.0, v[ii])
            self.assertGreaterEqual(1.0, v[ii])
        



"""Implements certain functions for solving an example of the Wahba problem.

Reference:

  [1] Valenti, Dryanovski, Xiao. 2015. Keeping a good attitude: A
      quaternion-based orientation filter for IMUs and MARGs. Sensors
      15(8): 19302-30.

"""

import numpy as np
from math import sqrt

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyquat._pyquat import Quat

    from pyquat._pyquat import valenti_q_mag as q_mag
    from pyquat._pyquat import valenti_q_acc as q_acc
    from pyquat._pyquat import valenti_dq_mag as dq_mag
    from pyquat._pyquat import valenti_dq_acc as dq_acc
else:
    from _pyquat import Quat
    from _pyquat import valenti_q_mag as q_mag
    from _pyquat import valenti_q_acc as q_acc
    from _pyquat import valenti_dq_mag as dq_mag
    from _pyquat import valenti_dq_acc as dq_acc

    
def q_global_to_local(
        a: np.ndarray[Any, np.dtype[np.float64]],
        b: np.ndarray[Any, np.dtype[np.float64]]
    ) -> Quat:
    """Compute a global to local transformation using a relatively
    low-noise measurement a (e.g. from an accelerometer) and a
    relatively noisy measurement b (e.g. from a magnetometer).

    Reference:
        [1] Valenti, Dryanovski, Xiao. 2015. Keeping a good
            attitude: A quaternion-based orientation filter for IMUs and
            MARGs. Sensors 15(8): 19302-30.

    Args:
        a:  relatively noiseless accelerometer measurement (unit norm)
        b:  relatively noisy magnetometer measurement (unit norm)

    Returns:
        A quaternion \f$ q_G^L = q_{acc} \otimes q_{mag} \f$ mapping
        from the global frame G to the local sensor frame L. Note that
        G is an NED frame where D is aligned with gravity, and N points
        toward magnetic north (horizontal component), not true north.
    """

    q_a      = q_acc(a)
    # q_a maps from L to intermediate.
    # Thus q_a_conj maps from intermediate to L, where intermediate
    # is just an about-z rotation from G.
    # Transform vector b into the intermediate frame.
    l        = q_a.conjugated().rotate(b)
    
    # Find the quaternion q_mag which rotates l into a vector lying on
    # the northeast/northwest half-plane.
    q_m      = q_mag(l)

    # Return the mapping from global to local.
    return q_a * q_m



"""This module implements Davenport's q-Method [0] (outlined in [1]),
as described and notated in [2] (which is where the equation numbering
appearing in comments originates).

References:

[0] Davenport, P. 1968.

[1] De Ruiter, A.; Damaren, C.; Forbes, J. 2013. Spacecraft Dynamics
    and Control: An Introduction, 1st Ed. Wiley, West Sussex,
    U.K. pp. 468-471.

[2] Ainscough, T.; Zanetti, R.; Christian, J.; Spanos, P. D. 2015.
    Q-Method extended Kalman filter. Journal of Guidance, Control, and
    Dynamics 38(4): 752-760.

"""
from typing import Any, Optional, Sequence, TYPE_CHECKING, cast

import numpy as np
import scipy.linalg as spl

if TYPE_CHECKING:
    import pyquat._pyquat as _pq
    from wahba import attitude_profile_matrix
    from wahba import davenport_matrix
else:
    import _pyquat as _pq
    from pyquat.wahba import attitude_profile_matrix
    from pyquat.wahba import davenport_matrix

def qekf_measurement_model(
        T: np.ndarray[Any, np.dtype[np.float64]],
        y: np.ndarray[Any, np.dtype[np.float64]],
        n: np.ndarray[Any, np.dtype[np.float64]],
        w: np.ndarray[Any, np.dtype[np.float64]]
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Generate a weighted least squares measurement model.

    This is eq. 18 from [2].

    Args:
        T   3x3 rotation matrix describing the prior attitude
        y   3xm matrix of unit 3D vector observations
        n   3xm matrix of corresponding unit 3D reference vectors
        w   list of weights corresponding to y and n

    Returns:
        A 3x3 measurement Jacobian.
    """
    H = np.zeros((3, 3))

    for ii in range(y.shape[1]):

        # Measurement Jacobian: eq 18
        yx  = _pq.skew(y[0:3,ii])
        Tn  = T.dot(n[0:3,ii])
        Tnx = _pq.skew(Tn)
        H += w[ii] * (yx.dot(Tnx) + Tnx.dot(yx))
        
    return H


def quest_measurement_covariance(
        vector: np.ndarray[Any, np.dtype[np.float64]],
        sigma: np.float64,
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Generate the measurement covariance from the QUEST measurement
    model for a given 3D unit vector and sigma.

    These are eqs. 5(a) and 5(b) in [2].

    Args:
        vector  unit vector
        sigma   standard deviation

    Returns:
        A 3x3 covariance matrix.
    """
    
    cov = (np.identity(3) - vector[0:3].reshape((3, 1)).dot(vector[0:3].reshape((1,3)))) * sigma

    # This may be the wrong way to get around this mypy error. The issue is
    # that numpy typing is funky, and the cast() function is the only way
    # I could find to make it work.
    return cast(np.ndarray[Any, np.dtype[np.float64]], cov)

def qekf_measurement_covariance(
        T: np.ndarray[Any, np.dtype[np.float64]],
        y: np.ndarray[Any, np.dtype[np.float64]],
        n: np.ndarray[Any, np.dtype[np.float64]],
        w: np.ndarray[Any, np.dtype[np.float64]],
        sigma_y: np.ndarray[Any, np.dtype[np.float64]],
        sigma_n: np.ndarray[Any, np.dtype[np.float64]]
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Generate a measurement covariance for the z vector by computing
    measurement covariances on the observations and reference vectors.

    The z vector is defined as the weighted sum of the cross-products
    of the observations and references. If

      B = \sum_{i=1}^n w_i y_i n_i^\top

    then

      \left[ z \times \right] = B^\top - B

    or alternatively,

      z = \sum_{i=1}^n w_i ( y_i \times n_i ) .

    This function first generates QUEST measurement model covariances
    for n and y (Eqs. 5(a) and 5(b) from [2]) and then uses these to
    produce a measurement covariance for z (Eq. 25 from [2]).

    The weights can be obtained using compute_weights(), or you can
    supply your own.

    Args:
        T        transformation matrix describing the prior attitude
        y        3xm matrix of unit vector observations
        n        3xm matrix of corresponding unit reference vectors
        w        array of measurement weights corresponding to y and n
        sigma_y  array of standard deviations for each unit vector observation
        sigma_n  array of standard deviations for each unit reference vector

    Returns:
        A weighted 3x3 measurement covariance.

    """
    R = np.zeros((3, 3))
    
    for ii in range(y.shape[1]):
        # QUEST measurement model: eq 5a, 5b
        Rnn = quest_measurement_covariance(n[:,ii], sigma_n[ii])
        Ryy = quest_measurement_covariance(y[:,ii], sigma_y[ii])

        yx  = _pq.skew(y[0:3,ii])
        Tn  = T.dot(n[0:3,ii])
        Tnx = _pq.skew(Tn)
        
        # Measurement covariance: eq 25
        yxT = yx.dot(T)
        R += (yxT.dot(Rnn.dot(yxT.T)) + Tnx.dot(Ryy.dot(Tnx.T))) * (4 * w[ii]**2)

    return R

        
def compute_weights(
        sigma_y: np.ndarray[Any, np.dtype[np.float64]],
        sigma_n: np.ndarray[Any, np.dtype[np.float64]]
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Given the sigmas for observations y and reference vectors n, this
    method computes the weights needed for the measurement covariance for
    z.

    This method represents Eq. 6 from [2].

    Args:
        sigma_y  standard deviations for each observation vector y
        sigma_n  standard deviations for each reference vector n

    Returns:
        A numpy array containing as many entries as there are
    observation and reference vectors.

    """
    w = []
    for ii in range(len(sigma_y)):
        # Compute weights: eq 6
        w.append( 1.0 / (sigma_n[ii]**2 + sigma_y[ii]**2) )
    return np.array(w)

        
def qmethod(
        y: np.ndarray[Any, np.dtype[np.float64]],
        n: np.ndarray[Any, np.dtype[np.float64]],
        w: np.ndarray[Any, np.dtype[np.float64]],
        q_prior: Optional[_pq.Quat] = None,
        N_prior: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None
    ) -> _pq.Quat:
    """Runs Davenport's q-Method as described in [2], using priors as
    needed for a SOAR filter or qEKF. It produces a normalized quaternion
    corresponding to the largest eigenvalue of the Davenport matrix.

    See p. 5 in [2] for additional details.

    You can provide either weights w or sigmas sigma_y and sigma_n.
    Weights are computed from the latter two using compute_weights()
    if weights are not supplied.

    Args:
        y        3xm matrix of 3D unit vector observations
        n        3xm matrix of corresponding 3D unit reference vectors 

    Kwargs:
        w        weights corresponding to each of the m entries of y 
                 and n; can also be computed from sigma_y and sigma_n
                 if those are supplied instead
        q_prior  prior attitude, if any
        N_prior  prior information matrix (3x3); this is also the
                 inverse of the prior covariance matrix (you must
                 provide this if you provide q_prior)

    Returns:
        A normalized quaternion.

    """
    if q_prior is None:
        T = np.identity(3)
        q_prior = _pq.identity()

        if N_prior is not None:
            raise ValueError("expected prior attitude to be given with prior information")
    else:
        T = q_prior.to_matrix()

        if N_prior is None:
            raise ValueError("expected prior information to be given with prior attitude")

    B = attitude_profile_matrix(obs = y, ref = n, weights = w)
    K = davenport_matrix(B, covariance_analysis = True)

    if N_prior is not None:
        # A0 is twice the inverse covariance.
        A0 = N_prior * 2

        # "Condition" K on the prior attitude and prior information
        Xi_prior = q_prior.big_xi()
        K -= Xi_prior.dot(A0.dot(Xi_prior.T))        

    # Get eigenvalues and eigenvectors
    lam, v = spl.eig(K)
    ii = np.argmax(lam) # find the largest eigenvalue

    return _pq.Quat(*v[:,ii]).normalized()



from typing import Union, Optional, Any, Tuple, Sequence, Callable, overload
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyquat._pyquat import *
else:
    from _pyquat import *
    
import math
import numpy as np
import numpy.typing as npt

from scipy import linalg
import warnings

QUAT_SMALL = 1e-8

def fromstring(*args, **kwargs) -> Quat:
    """
    Shortcut for pyquat.Quat.from_vector(numpy.fromstring()).  If you
    don't provide a 'sep' argument, this method will supply the
    argument count=4 to numpy.fromstring() regardless of what you
    provided for it.
    """
    if 'sep' in kwargs and kwargs['sep'] == '':
        kwargs['count'] = 4
        
    return Quat(*(np.fromstring(*args, **kwargs)))

def qdot(
        q: Union[
            Quat,
            np.ndarray[Any, np.dtype[np.float64]]
        ],
        w: np.ndarray[Any, np.dtype[np.float64]],
        big_w: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Compute dq/dt given some angular velocity w and initial quaternion q.
    """
    if isinstance(q, Quat):
        q = q.to_vector()
    if big_w is None:
        big_w = big_omega(w)
    return np.dot(big_w * 0.5, q)


def wdot(
        w: np.ndarray[Any, np.dtype[np.float64]],
        J: np.ndarray[Any, np.dtype[np.float64]],
        J_inv: Optional[npt.ArrayLike] = None
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Compute dw/dt given some angular velocity w and moment of inertia J.
    """
    if J_inv is None:
        J_inv = linalg.inv(J)
    return np.dot(J_inv, np.dot(skew(np.dot(J, w)), w))

def state_transition_matrix(
        w: np.ndarray[Any, np.dtype[np.float64]],
        big_w: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Generate a state transition matrix for a quaternion based on some
    angular velocity w.
    """
    if big_w is None:
        big_w = big_omega(w)
    return big_w * 0.5

def change(*args, **kwargs):
    warnings.warn("deprecated", DeprecationWarning)
    return propagate(*args, **kwargs)

def propagate(
        q: Quat,
        w: np.ndarray[Any, np.dtype[np.float64]],
        dt: float):
    """
    Change a quaternion q by some angular velocity w over some small
    timestep dt.
    """

    # Find magnitude of angular velocity (in r/s)
    w_norm = linalg.norm(w)
    if w_norm < QUAT_SMALL:
        return q.copy()
    return Quat(*(np.dot(expm(w, dt), q.to_vector())))

def matrix_propagate(
        T: np.ndarray[Any, np.dtype[np.float64]],
        w: np.ndarray[Any, np.dtype[np.float64]],
        dt: float,
        r: int = 1
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Propagate an attitude matrix T forward by some angular velocity w
    over time step dt. This method uses a Taylor expansion of degree r
    where r is between 1 and 4 inclusive.

    Note that there's minimal computational difference between orders
    2 and 3, as each involves a second 3x3 matrix multiplication. The
    step to 4th order is also quite small, involving only an
    additional vector dot product. In most cases, you will probably
    want to use r = 1 or r = 4.

    Args:
        T:  transformation matrix (3x3)
        w:  angular velocity vector (length 3)
        dt: time step size
        r:  Taylor expansion degree (between 1 and 4 inclusive)

    Returns:
        A 3x3 matrix giving the updated transformation.
    """

    wt   = w*dt
    wtx  = skew(wt)
    
    exp  = np.identity(3) + wtx
    if r >= 2:
        wtx2 = wtx.dot(wtx)
        if r == 2:
            exp += wtx2 * 0.5
        elif r >= 3:
            if r == 3:
                exp += wtx2 * (0.5 - wt / 6.0)
            elif r == 4:
                wt2   = wt.T.dot(wt)
                exp += wtx2 * (0.5 - wt / 6.0 - wt2 / 24.0)
            else:
                raise NotImplementedError("degree must be between 1 and 4 inclusive")
        else:
            raise NotImplementedError("degree must be between 1 and 4 inclusive")

    return exp.T.dot(T)
        
    

def propagate_additively(
        q: Quat,
        w: np.ndarray[Any, np.dtype[np.float64]],
        dt: float
    ) -> Quat:
    """Change a quaternion q by some angular velocity w over some small
    timestep dt, using additive propagation (q1 = q0 + dq/dt * dt)"""
    
    q_vector = q.to_vector()
    q_vector += qdot(q_vector, w) * dt
    return Quat(*q_vector)

    
def cov(
        ary: Union[
            np.ndarray[Any, np.dtype[np.float64]],
            np.ndarray[Any, np.dtype[np.object_]]
        ]
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Compute the covariance of an array of quaternions, where each
    column represents a quaternion.
    """
    # If the user supplies an array of N quaternions, convert it to a 4xN array,
    # since we need it in this form to get its covariance.
    if ary.dtype == np.dtype(Quat):
        a = np.empty((4, max(ary.shape)), dtype=np.float64)
        q_ary = ary.T
        for i, q in enumerate(q_ary.flatten()):
            a[:,i] = q.to_vector()[:,0]
        ary = a
            
    # Compute the covariance of the supplied quaternions.
    return np.cov(ary)

def mean(
        ary: npt.ArrayLike,
        covariance: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None
    ) -> Quat:
    """
    Compute the average quaternion using Markey, Cheng, Craissidis, and Oshman (2007)
    
    This method takes a 4xN array and computes the average using eigenvalue decomposition.
    """
    if covariance == None:
        covariance = cov(ary)

    # Compute their eigenvalues and eigenvectors
    eigenvalues, eigenvectors = linalg.eig(covariance)
    max_index = np.argmax(eigenvalues)
    q = eigenvectors[max_index]
    mean = Quat(q[0], q[1], q[2], q[3])
    mean.normalize()
    return mean

def mean_and_cov(
        ary: npt.ArrayLike
    ) -> Tuple[Quat, np.ndarray[Any, np.dtype[np.float64]]]:
    c = cov(ary)
    m = mean(ary, covariance=c)
    return (m,c)

def angle_vector_cov(
        ary: Union[
            np.ndarray[Any, np.dtype[np.float64]],
            np.ndarray[Any, np.dtype[np.object_]]
        ]
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Compute the covariance of an array of quaternions, like cov(), except use the attitude vector
    representation of each.
    """
    if ary.dtype == np.dtype(Quat):
        a = np.empty((3, ary.shape[1]), dtype=np.float64)
        q_ary = ary.T
        for i, q in enumerate(q_ary):
            a[:,i] = Quat(q[0], q[1], q[2], q[3]).to_angle_vector()[:,0]
        ary = a
    elif ary.dtype == np.float64:
        a = np.empty((3, max(ary.shape)), dtype=np.float64)
        q_ary = ary.T
        for i, q in enumerate(q_ary.flatten()):
            a[:,i] = q.to_angle_vector()[:,0]
        ary = a
    else:
        raise TypeError("expected Quat or np.float64 dtype")

    return np.cov(ary)

def from_rotation_vector(v: np.ndarray[Any, np.dtype[np.float64]]) -> Quat:
    """
    Shortcut for Quat.from_rotation_vector(v).
    """
    return Quat.from_rotation_vector(v)

def from_matrix(m: np.ndarray[Any, np.dtype[np.float64]]) -> Quat:
    """
    Shortcut for Quat.from_matrix(v).
    """
    return Quat.from_matrix(m)


"""
        q: Union[Quat, npt.ArrayLike],
        w: npt.ArrayLike,
        big_w: Optional[npt.ArrayLike] = None
    ) -> np.ndarray[Any, np.dtype[np.float64]
               """     

def step_rk4(
        q: Quat,
        w: np.ndarray[Any, np.dtype[np.float64]],
        dt: float,
        w_dynamics: Optional[Callable[[npt.ArrayLike, npt.ArrayLike, npt.ArrayLike],np.ndarray[Any, np.dtype[np.float64]]]] = None,
        q_dynamics: Callable[
            [
                np.ndarray[Any, np.dtype[np.float64]],
                np.ndarray[Any, np.dtype[np.float64]]
            ], np.ndarray[Any, np.dtype[np.float64]]] = qdot,
        J: Optional[npt.ArrayLike] = None,
        J_inv: Optional[npt.ArrayLike] = None
    ) -> Tuple[Quat, np.ndarray[Any, np.dtype[np.float64]]]:
    """
    Use a standard Runge-Kutta 4-step / 4th-order integration to step
    the quaternion forward in time.
    """
    if linalg.norm(w) <= QUAT_SMALL:
        return (q, w)
    
    #C1 = 0.0
    #C2 = 0.5
    #C3 = 0.5
    #C4 = 1.0
    A21 = 0.5
    A32 = 0.5 #A31 = 0.0
    #A43 = 1.0; A42 = A41 = 0.0
    #B1 = 1/6.0
    #B2 = 1/3.0
    #B3 = B2
    #B4 = B1

    q1  = q.to_vector()
    
    if w_dynamics is None:
        qk1 = q_dynamics(q1, w)
        q2  = q1 + qk1 * A21 * dt
        qk2 = q_dynamics(q2, w)
        q3  = q1 + qk2 * A32 * dt # + qk1 * A31 * dt
        qk3 = q_dynamics(q3, w)
        q4  = q1 + qk3 * dt # + A42 * qk2 * dt + A41 * qk1 * dt
        qk4 = q_dynamics(q4, w)

        w_next = np.array(w)
    else:
        if J is None:
            J = np.identity(3)
            J_inv = J
        elif J_inv is None:
            J_inv = linalg.inv(J)
                 
        w1  = w # + C1 * dt
        wk1 = w_dynamics(w1, J, J_inv)
        qk1 = q_dynamics(q1, w1) #q = q1, w = w1

        q2  = q1 + qk1 * A21 * dt
        w2  = w  + wk1 * A21 * dt
        wk2 = w_dynamics(w2, J, J_inv)
        qk2 = q_dynamics(q2, w2)


        q3  = q1 + qk2 * A32 * dt # + qk1 * A31 * dt
        w3  = w  + wk2 * A32 * dt # + wk1 * A31 * dt
        wk3 = w_dynamics(w3, J, J_inv)
        qk3 = q_dynamics(q3, w3)

        q4  = q1 + qk3 * dt # + A42 * qk2 * dt + A41 * qk1 * dt
        w4  = w  + wk3 * dt # same
        wk4 = w_dynamics(w4, J, J_inv)
        qk4 = q_dynamics(q4, w4)
        
        w_next = w1 + dt * (wk1 + wk2*2 + wk3*2 + wk4) / 6.0

    q_next = q1 + dt * (qk1 + qk2*2 + qk3*2 + qk4) / 6.0
    
    return (Quat(*q_next).normalized(), w_next)

def step_cg3(
        q: Quat,
        w: np.ndarray[Any, np.dtype[np.float64]],
        dt: float,
        w_dynamics: Optional[Callable[[npt.ArrayLike, npt.ArrayLike, npt.ArrayLike],np.ndarray[Any, np.dtype[np.float64]]]] = None,
        J: Optional[npt.ArrayLike] = None,
        J_inv: Optional[npt.ArrayLike] = None
    ) -> Tuple[Quat, np.ndarray[Any, np.dtype[np.float64]]]:
    """
    Use a 3-stage, third-order Crouch-Grossman integration for 
    propagating a quaternion and a Runge-Kutta integration for 
    propagating angular velocity.

    This method returns a tuple containing the resulting quaternion
    and omega.
    """
    if linalg.norm(w) <= QUAT_SMALL:
        return (q, w)
    
    B1 = 13/51.0
    B2 = -2/3.0
    B3 = 24/17.0
    A21 = 0.75
    A31 = 119/216.0
    A32 = 17/108.0
    C1 = 0.0
    C2 = 3/4.0
    C3 = 17/24.0

    q1  = q.to_vector()
    
    if w_dynamics is None:
        w_next = np.array(w)
        
        expm3 = expm(w, dt * B3)
        expm2 = expm(w, dt * B2)
        expm1 = expm(w, dt * B1)        

    else:
        if J is None:
            J = np.identity(3)
            J_inv = J
        elif J_inv is None:
            J_inv = linalg.inv(J)
        
        w1  = w
        wk1 = w_dynamics(w1, J, J_inv)

        w2  = w1 + A21 * wk1 * dt
        wk2 = w_dynamics(w2, J, J_inv)

        w3  = w1 + (wk1 * A31 + wk2 * A32) * dt
        wk3 = w_dynamics(w3, J, J_inv)

        expm3 = expm(w3, dt * B3)
        expm2 = expm(w2, dt * B2)
        expm1 = expm(w1, dt * B1)

        w_next = w + (wk1*B1 + wk2*B2 + wk3*B3) * dt
    
    q_next = np.dot(expm3, np.dot(expm2, np.dot(expm1, q1)))

    return (Quat(*q_next), w_next)

def step_cg4(
        q: Quat,
        w: np.ndarray[Any, np.dtype[np.float64]],
        dt: float,
        w_dynamics: Optional[Callable[[npt.ArrayLike, npt.ArrayLike, npt.ArrayLike],np.ndarray[Any, np.dtype[np.float64]]]] = None,
        J: Optional[npt.ArrayLike] = None,
        J_inv: Optional[npt.ArrayLike] = None
    ) -> Tuple[Quat, np.ndarray[Any, np.dtype[np.float64]]]:
    """
    Use a 5-stage, fourth-order Crouch-Grossman integration for 
    propagating a quaternion and a Runge-Kutta integration for 
    propagating angular velocity.

    This method returns a tuple containing the resulting quaternion
    and omega.
    """
    if linalg.norm(w) <= QUAT_SMALL:
        return (q, w)
    
    B1 =   0.1370831520630755
    B2 =  -0.0183698531564020
    B3 =   0.7397813985370780
    B4 =  -0.1907142565505889
    B5 =   0.3322195591068374
    A21 =  0.8177227988124852
    A31 =  0.3199876375476427
    A32 =  0.0659864263556022
    A41 =  0.9214417194464946
    A42 =  0.4997857776773573
    A43 = -1.0969984448371582
    A51 =  0.3552358559023322
    A52 =  0.2390958372307326
    A53 =  1.3918565724203246
    A54 = -1.1092979392113565
    C1  =  0.0
    C2  =  0.8177227988124852
    C3  =  0.3859740639032449
    C4  =  0.3242290522866937
    C5  =  0.8768903263420429

    q1  = q.to_vector()
    
    if w_dynamics is None: # don't propagate omega
        expm5 = expm(w, dt * B5)
        expm4 = expm(w, dt * B4)
        expm3 = expm(w, dt * B3)
        expm2 = expm(w, dt * B2)
        expm1 = expm(w, dt * B1)
    
        w_next = np.array(w)
        
    else:
        if J is None:
            J = np.identity(3)
            J_inv = J
        elif J_inv is None:
            J_inv = linalg.inv(J)

        w1  = w
        wk1 = w_dynamics(w1, J, J_inv)

        w2  = w1 + A21 * wk1 * dt
        wk2 = w_dynamics(w2, J, J_inv)

        w3  = w1 + (wk1 * A31 + wk2 * A32) * dt
        wk3 = w_dynamics(w3, J, J_inv)

        w4  = w1 + (wk1 * A41 + wk2 * A42 + wk3 * A43) * dt
        wk4 = w_dynamics(w4, J, J_inv)
        
        w5  = w1 + (wk1 * A51 + wk2 * A52 + wk3 * A53 + wk4 * A54) * dt
        wk5 = w_dynamics(w5, J, J_inv)

        expm5 = expm(w5, dt * B5)
        expm4 = expm(w4, dt * B4)
        expm3 = expm(w3, dt * B3)
        expm2 = expm(w2, dt * B2)
        expm1 = expm(w1, dt * B1)
    
        w_next = w + (wk1*B1 + wk2*B2 + wk3*B3 + wk4*B4 + wk5*B5) * dt

    # propagate the quaternion
    q_next = np.dot(expm5,
                    np.dot(expm4,
                           np.dot(expm3,
                                  np.dot(expm2,
                                         np.dot(expm1, q1)))))


    return (Quat(*q_next), w_next)


def interp(
        x: Sequence[float],
        xp: Sequence[float],
        fp: Sequence[Quat],
        lerp_threshold: float = 1.0,
        left: str = 'slerp',
        right: str = 'slerp'
    ) -> Sequence[Quat]:
    """Corresponds roughly to numpy.interp. Attempts to interpolate the
    quaternion series qp occurring at independent variable series tp (e.g.
    times) at other values given by t.

    Note that left and right, when set to 'slerp', ignore lerp_threshold.

    Args:
        x:               time or times at which to interpolate
        xp:              times corresponding to qp values
        fp:              quaternion series
        lerp_threshold:  dot product threshold for deciding between lerp
                         and slerp (default is 1.0, always uses slerp; for
                         always use lerp, choose 0.0)
        left:            'lerp' or 'slerp' (default 'slerp')
        right:           'lerp' or 'slerp' (default 'slerp')

    Returns:
        If x is a single value, returns a single quaternion. Otherwise,
        returns a list of quaternions of same size as x.
    """

    indices = np.searchsorted(xp, x)
    f       = []

    jj = 0
    for ii in indices:
        if ii == 0: # <= xp (left)
            t = (x[jj] - xp[0]) / (xp[1] - xp[0])
            if left == 'slerp':
                f.append(fp[0].slerp(fp[1], t, lerp_threshold = 1.0))
            else:
                f.append(fp[0].lerp(fp[1], t))

        elif ii == len(fp): # >= xp (right)
            t = (x[jj] - xp[ii-2]) / (xp[ii-1] - xp[ii-2])
            if right == 'slerp':
                f.append(fp[ii-2].slerp(fp[ii-1], t, lerp_threshold = 1.0))
            else:
                f.append(fp[ii-2].lerp(fp[ii-1], t))

        else: # all others
            t = (x[jj] - xp[ii-1]) / (xp[ii] - xp[ii-1])
            f.append(fp[ii-1].slerp(fp[ii], t, lerp_threshold = lerp_threshold))

        jj += 1

    return f



from typing import overload, Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pyquat._pyquat import Quat
else:
    from _pyquat import Quat

import numpy as np

@overload
def attitude_profile_matrix(
    q: Quat,
    cov: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
    inverse_cov: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None
) -> np.ndarray[Any, np.dtype[np.float64]]:
    ...

@overload
def attitude_profile_matrix(
    obs: np.ndarray[Any, np.dtype[np.float64]],
    ref: np.ndarray[Any, np.dtype[np.float64]],
    weights: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None
) -> np.ndarray[Any, np.dtype[np.float64]]:
    ...


import typing
from typing import TYPE_CHECKING

import numpy as np
import numpy.random as npr
import scipy.linalg as spl

from test.assertions import QuaternionTest

if TYPE_CHECKING:
    import pyquat.pyquat as pq
    import pyquat.pyquat.wahba.qmethod as qmethod
else:
    import pyquat as pq
    import pyquat.wahba.qmethod as qmethod

class TestWahbaQMethod(QuaternionTest):
    
    def test_qmethod(self) -> None:
        q_inrtl_to_body_list = [
            pq.identity(),
            pq.Quat(1.0, -2.0, 3.0, 4.0).normalized(),
            pq.Quat(0.0, 1.0, 0.0, 0.0),
            pq.Quat(0.0, 0.0, 1.0, 0.0),
            pq.Quat(0.0, 0.0, 0.0, 1.0)]

        # Setup misalignments using skew matrices.
    
        # Our truth will be corrupted to reference by some error in our
        # knowledge about the world.

        # We will then corrupt those reference vectors by observation
        # errors (because our instruments are kind of shitty).
    
        ref_misalign     = npr.randn(3) * 1e-6
        sun_obs_misalign = npr.randn(3) * 1e-5
        mag_obs_misalign = npr.randn(3) * 1e-5
    
        T_ref_err = np.identity(3) #- pq.skew(ref_misalign)
        T_sun_obs_err = np.identity(3) - pq.skew(sun_obs_misalign)
        T_mag_obs_err = np.identity(3) - pq.skew(mag_obs_misalign)
    
        mag_truth = np.array([0.0, 0.1, 1.0])
        mag_truth /= spl.norm(mag_truth)

        sun_truth = np.array([0.5, 0.5, 0.02])
        sun_truth /= spl.norm(sun_truth)

        for qib in q_inrtl_to_body_list:
            Tib = qib.to_matrix()

            mag_ref = T_ref_err.dot(mag_truth)
            mag_ref /= spl.norm(mag_ref)
            sun_ref = T_ref_err.dot(sun_truth)
            sun_ref /= spl.norm(sun_ref)

            mag_obs  = T_mag_obs_err.dot(Tib.dot(mag_ref))
            mag_obs /= spl.norm(mag_obs)
            sun_obs  = T_sun_obs_err.dot(Tib.dot(sun_ref))
            sun_obs /= spl.norm(sun_obs)

            P_prior = np.identity(3) * np.pi**2
            N_prior = spl.inv(P_prior)


            # Assemble arguments to pass to qmethod()
            sigma_y = [1e-2, 1e-2]
            sigma_n = [1e-2, 1e-2]
            weights = qmethod.compute_weights(sigma_y, sigma_n)
            qmethod_args = (np.vstack((sun_obs, mag_obs)).T,
                            np.vstack((sun_ref, mag_ref)).T,
                            weights)            

            # Test qmethod with priors
            q_posterior = qmethod.qmethod(*qmethod_args,
                                          q_prior = pq.identity(),
                                          N_prior = N_prior)

            # Test qmethod without priors
            q_naive     = qmethod.qmethod(*qmethod_args)

            for q in (q_posterior, q_naive):
                dq_body = qib * q.conjugated()
                self.assertLess(np.abs(np.arccos(dq_body.w)), 1e-4)



import typing
from typing import TYPE_CHECKING

import numpy as np
from scipy import linalg

from test.assertions import QuaternionTest
        
if TYPE_CHECKING:
    import pyquat.pyquat.wahba.valenti as valenti
    import pyquat.pyquat.random as random
else:
    import pyquat.wahba.valenti as valenti
    import pyquat.random as random


class TestWahbaValenti(QuaternionTest):
    def test_q_acc_points_down(self) -> None:
        """ Tests that q_acc produces an arbitrary quaternion which rotates the gravity vector [0,0,1]' (down) to point at a measurement"""
        for ii in range(0,100):
            a0 = random.uniform_random_axis().reshape(3)
            d  = np.array([0.0, 0.0, 1.0])
        
            q  = valenti.q_acc(a0)
            a1 = q.rotate(d)
            np.testing.assert_almost_equal(a0, a1, decimal=12)

            q  = valenti.dq_acc(a0)
            a1 = q.rotate(d)
            np.testing.assert_almost_equal(a0, a1, decimal=12)

    def test_q_mag_points_north(self) -> None:
        """ Tests that q_mag produces an arbitrary quaternion which points a random vector at [1,0,0]' (magnetic north); also tests that q_global_to_local gives the same result"""
        for ii in range(0, 100):
            a  = random.uniform_random_axis().reshape(3)
            b  = random.uniform_random_axis().reshape(3)
        
            while a.dot(b) > 0.95: # make sure b isn't the same direction as a
                b = random.uniform_random_axis().reshape(3)
        
            qa = valenti.q_acc(a) # qa rotates a to being 0,0,1 (so into an xyD frame)
            l  = qa.conjugated().rotate(b)
            qm = valenti.q_mag(l)

            dqm = valenti.dq_mag(l)

            # Test that qm is about-z only
            self.assertEqual(qm.x, 0.0)
            self.assertEqual(qm.y, 0.0)

            # Do the same for dq_mag
            self.assertEqual(dqm.x, 0.0)
            self.assertEqual(dqm.y, 0.0)
        
            q_G_to_L  = qa * qm
            q_L_to_G  = q_G_to_L.conjugated()

            # Test that q_G_to_L still points gravity at a measurement
            # (really actually unnecessary since qm is about-z only)
            d         = np.array([0.0, 0.0, 1.0])
            a1        = q_G_to_L.rotate(d)
            np.testing.assert_almost_equal(a, a1, decimal=14)

            # Test that q_L_to_G rotates b so its horizontal component is
            # toward [1,0,0]'
            northish = q_G_to_L.conjugated().rotate(b)
            self.assertAlmostEqual(northish[1], 0.0, places=14)
            self.assertGreaterEqual(northish[0], 0.0)

            # Test that q_global_to_local gives the same result.
            q_G_to_L_0 = valenti.q_global_to_local(a, b)
            self.assert_equal(q_G_to_L, q_G_to_L_0)



"""
This file contains quaternion and attitude utility functions
relating to the ESOQ2 solution to what is known as Wahba's problem
[0], namely the problem of finding a rotation matrix which brings one
set of vectors into the best possible alignment with a second set of
vectors.

Reference:
[0] Wahba, Grace. 1965. Problems and Solutions: Problem 65--1.
    SIAM Review 7(3): 409.

"""
from typing import Any, Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from pyquat._pyquat import Quat
else:
    from _pyquat import Quat

import numpy as np
from scipy import linalg
from math import sqrt, pi, cos, acos

def trace_adj(
        a: np.ndarray[Any, np.dtype[np.float64]]
    ) -> np.float64:

    """Returns the trace of the adjunct of matrix a, which must be 3x3 or 4x4."""
    if max(a.shape) == 3:
        c1 = a[1,1] * a[2,2] - a[1,2] * a[2,1]
        c2 = a[0,0] * a[2,2] - a[0,2] * a[2,0]
        c3 = a[0,0] * a[1,1] - a[0,1] * a[1,0]
        return np.float64(c1 + c2 + c3)
    else:
        return minor_det_4x4(a, 0, 1, 2) + minor_det_4x4(a, 0, 2, 3) + minor_det_4x4(a, 1, 2, 3) + minor_det_4x4(a, 0, 1, 3)

def minor_det_4x4(
        a: np.ndarray[Any, np.dtype[np.float64]],
        i: int = 0,
        j: int = 1,
        k: int = 2
    ) -> np.float64:
    """Returns the determinant of a minor of a 4x4 matrix.

    Args:
        a      matrix whose minor we're interested in
        i,j,k  rows/columns to include in the determinant calculation (e.g.,
               if 0, 1, and 2 are supplied, the determinant will be of the
               upper left-hand submatrix)

    Returns:
        A Float giving the determinant of the indicated 3x3 submatrix.
        
    """
    return np.float64(a[i,i] * (a[j,j]*a[k,k] - a[j,k]*a[k,j]) - a[i,j] * (a[j,i]*a[k,k] - a[j,k]*a[k,i]) + a[i,k] * (a[j,i]*a[k,j] - a[j,j]*a[k,i]))

def trace_adj_symm(
        a: np.ndarray[Any, np.dtype[np.float64]]
    ) -> np.float64:

    """Returns the trace of the adjunct of a symmetric matrix a, which
    must be 3x3 or 4x4."""
    if max(a.shape) == 3:
        c1 = a[1,1] * a[2,2] - a[1,2]**2
        c2 = a[0,0] * a[2,2] - a[0,2]**2
        c3 = a[0,0] * a[1,1] - a[0,1]**2
        return np.float64(c1 + c2 + c3)
    else:
        return trace_adj(a)


def lambda_nr_iterate(
        lambda_c: float,
        b: float,
        c: float,
        d: float
    ) -> float:
    l2 = lambda_c**2
    l3 = l2*lambda_c
    l4 = l2**2
    return lambda_c - (l4 + b*l2 + c*lambda_c + d) / (4.0*l3 + 2.0*b*lambda_c + c)

def sequential_rotation_helper(
        B: np.ndarray[Any, np.dtype[np.float64]] 
    ) -> int:
    """
    See documentation for sequential_rotation()
    """
    
    diag_B  = np.diag(B)
    irot    = int(np.argmin(diag_B))
    sequential_rotation(B, irot = irot)
    return irot


def quat_sequential_rotation(
        q: Quat,
        irot: int,
    ) -> Quat:
    """See documentation for sequential_rotation()

    """

    if irot == 0:
        return Quat(-q.x, q.w, -q.z, q.y)
    elif irot == 1:
        return Quat(-q.y, q.z, q.w, -q.x)
    elif irot == 2:
        return Quat(-q.x, q.w, -q.z, q.y)
    else:
        raise ValueError("expected irot to be 0, 1, or 2")

def sequential_rotation(
        B: np.ndarray[Any, np.dtype[np.float64]],
        irot: Optional[int] = None
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """"It is noted without proof that a rotation through an angle
    greater than \frac{\pi}{2} can be expressed as a rotation
    through \pi about one of the coordinate axes followed by a
    rotation about a new axis through an angle less than 
    \frac{\pi}{2}. An initial rotation through \pi about one of
    the coordinate axes is equivalent to changing the signs of
    two components of each of the reference vectors. The 
    quaternion p=(p_1,p_2,p_3,p_4)^\T of the optimal rotation
    transforming the new reference vectors \hat{V}_i', i=1,...,n
    into the observation vectors \hat{W}_i, i=1,...,n as
    calculated...is related very simply to the desired optimal
    quaternion. The results for the three possible cases are

    1. Initial rotation through \pi about the x axis:
        \hat{V}_i' = (\hat{V}_{ix}, -\hat{V}_{iy}, -hat{V}_iz)
        q          = (p_4, -p_3, p_2, -p_1)^\T
    2. Initial rotation through \pi about the y axis:
        \hat{V}_i' = (-\hat{V}_{ix}, \hat{V}_{iy}, -hat{V}_iz)
        q          = (p_3, p_4, -p_1, -p_2)^\T
    3. Initial rotation through \pi about the z axis:
        \hat{V}_i' = (-\hat{V}_{ix}, -\hat{V}_{iy}, hat{V}_iz)
        q          = (-p_2, p_1, p_4, -p_3)^\T
   
    Clearly, that initial rotation (including no initial rotation)
    will yield the most accurate estimate of q_\textrm{opt} for
    which |\gamma| achieves the largest value. In any practical
    application, however, |\gamma| can be allowed to become quite
    small before the method of sequential rotations need [sic.]
    to be invoked." [0]

    This method has three different calling modes:

    1. Supply irot and attitude profile matrix B. This performs the
       specified rotation on B and returns B.
    2. Supply irot and quaternion q. This performs the specified
       rotation on q and returns q.
    3. Supply only an attitude profile matrix B. Determines the
       rotation to perform, performs the rotation, and returns
       irot.

    This method produces an *in-place* sequential rotation of the
    attitude profile matrix B and returns the rotation performed. It
    is indifferent to the angle of rotation.

    Note: pyquat uses a different quaternion ordering than Shuster 
    & Oh.

    Reference:
    [0] Shuster, Oh. 1981. Three-axis attitude determination from 
        vector observations. Journal of Guidance and Control 4(1): 
        70-77.

    """

    if irot is None:
        irot = sequential_rotation_helper(B)
    if irot == 0:
        B[:,1:3] *= -1.0
    elif irot == 1:
        B[:,0] *= -1.0
        B[:,2] *= -1.0
    elif irot == 2:
        B[:,0:2] *= -1.0
    return B


def davenport_eigenvalues(
        K: np.ndarray[Any, np.dtype[np.float64]],
        B: np.ndarray[Any, np.dtype[np.float64]],
        n_obs: Optional[int] = None
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Compute the eigenvalues of the Davenport matrix K.

    Assumes sequential rotation has already been applied.

    Take careful note that you should pass n_obs = 2 if you only have
    two observations making up your attitude profile matrix B.

    Reference:

    [0] Mortari (1997). ESOQ-2 single-point algorithm for fast
        optimal spacecraft attitude determination. Advances in the
        Astronautical Sciences 95: 817-826.

    Args:
        K      Davenport matrix (4x4)
        B      attitude profile matrix (3x3)
        n_obs  number of observations involved in computing B (optional
               unless you only have 2, in which case this needs to be
               provided)

    Returns:
        The sorted eigenvalues of the Davenport K matrix.

    """
    tr_adj_K = trace_adj(K)
    tr_B     = np.trace(B)
    S        = B+B.T
    tr_adj_S = trace_adj_symm(S)
    z        = K[1:4,0].reshape((3,1))

    # a never actually gets used, so we won't compute it, but note
    # here that this is what it ought to be.

    # a = np.trace(K)
    b = -2.0 * tr_B**2 + tr_adj_S - np.dot(z.T, z)[0,0]
    d = linalg.det(K)

    if n_obs == 2:
        g3 = sqrt( 2*sqrt(d) - b)
        g4 = sqrt(-2*sqrt(d) - b)        
        lambda_4 =  (g3 + g4) / 2.0
        lambda_1 = -lambda_4
        lambda_3 =  (g3 - g4) / 2.0
        lambda_2 = -lambda_3
    else:
        c = -tr_adj_K

        p  = (b/3.0)**2 + 4*d / 3.0
        q  = (b/3.0)**3 - 4*d*b / 3.0 + c**2 / 2.0

        u1 = 2 * sqrt(p) * cos(acos(q / p**1.5) / 3.0) + b / 3.0

        g1 = sqrt(u1 - b)
        g2 = 2.0 * sqrt(u1**2 - 4.0*d)

        lambda_1 = (-g1 - sqrt(-u1 - b + g2)) / 2.0
        lambda_2 = (-g1 + sqrt(-u1 - b + g2)) / 2.0
        lambda_3 =  (g1 - sqrt(-u1 - b - g2)) / 2.0
        lambda_4 =  (g1 + sqrt(-u1 - b - g2)) / 2.0

    return np.array([lambda_4, lambda_3, lambda_2, lambda_1])


def esoq2(
        K: np.ndarray[Any, np.dtype[np.float64]],
        B: np.ndarray[Any, np.dtype[np.float64]],
        lambda_0: Optional[float] = None,
        n_obs: Optional[int] = None
    ) -> Tuple[Quat, float]:
    """
    Compute the optimal quaternion using Mortari's second estimator.
    
    References: 
    [0] Mortari (2000). Second estimator of the optimal quaternion. 
        Journal of Guidance, Control, and Dynamics 23(5): 885-8.
    [1] Mortari (1997). ESOQ: A closed-form solution to the Wahba
        problem. Journal of the Astronautical Sceinces 45(2): 195--204.
    [2] Mortari (1997). ESOQ-2 single-point algorithm for fast optimal
        spacecraft attitude determination. In Advances in the 
        Astronautical Sciences, 4--7 Aug, Sun Valley, Idaho, 
        pp. 817--826.

    Args:
        K          Davenport matrix (4x4)
        B          attitude profile matrix (3x3)
        lambda_0   optional initial guess about maximum eigenvalue (default
                   is to set this equal to n_obs)
        n_obs      optional number of observations involved in computing B

    Returns:
        A tuple of the optimal quaternion (corresponding to the maximum
    eigenvalue), which has been normalized; and the value of the loss
    function from Wahba's problem.

    """
    import pyquat as pq

    if lambda_0 is None: # Initial guess for maximum eigenvalue
        if n_obs is None:
            raise ValueError("expected n_obs to be provided if lambda_0 is not")
        lambda_0 = float(n_obs)

    tr_B = np.trace(B)
    z    = K[1:4,0].reshape((3,1))
    S    = B+B.T

    lambdas = davenport_eigenvalues(K, B, n_obs = n_obs)
    lambda_max = lambdas[0]
    loss = lambda_0 - lambda_max

    trace_plus_lambda  = tr_B + lambda_max
    trace_minus_lambda = tr_B - lambda_max
    
    S_minus_tpl = S - np.identity(3) * trace_plus_lambda
    M = S_minus_tpl * trace_minus_lambda - np.dot(z, z.T)
    m1 = M[0,0:3]
    m2 = M[1,0:3]
    m3 = M[2,0:3]

    # Find the largest e_i.
    e_options = [np.cross(m2, m3), np.cross(m3, m1), np.cross(m1, m2)]
    e_mags    = np.array([linalg.norm(e_options[0]), linalg.norm(e_options[1]), linalg.norm(e_options[2])])
    imax      = np.argmax(e_mags)
    e         = e_options[imax].reshape((3,1))

    if n_obs != 2:
        # This piece borrowed from:
        # https://github.com/muzhig/ESOQ2/blob/master/esoq2p1.py.
        # Has not yet been tested!!!
        n1 = np.array([S_minus_tpl[0,0] - 2 * lambda_max, S_minus_tpl[0,1], S_minus_tpl[2,0]])
        n2 = np.array([S_minus_tpl[0,1], S_minus_tpl[1,1] - 2 * lambda_max, S_minus_tpl[1,2]])
        n3 = np.array([S_minus_tpl[2,0], S_minus_tpl[1,2], S_minus_tpl[2,2] - 2 * lambda_max])
        aa = [m2, m3, m1][imax]
        bb = [n3, n1, n2][imax]
        cc = [m3, m1, m2][imax]
        dd = [n2, n3, n1][imax]
        mm = [m1, m2, m3][imax].reshape((1,3))
        nn = [n1, n2, n3][imax].reshape((1,3))

        v = (np.cross(aa, bb) - np.cross(cc, dd)).reshape((3,1))
        loss = - (mm.dot(e)) / (nn.dot(e) + mm.dot(v))
        trace_minus_lambda  += loss
        e += loss * v
        

    # Find the normalized quaternion (unrotated)
    q = Quat(np.dot(z.T, e),
             *(e * -trace_minus_lambda)).normalized_large()
        
    return q, loss



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from typing import TYPE_CHECKING, Any, Dict, Optional, List

if TYPE_CHECKING:
    from pyquat._pyquat import Quat
else:
    from _pyquat import Quat

def prepare_quaternion_array_for_plotting(
        q_ary: np.ndarray[Any, np.dtype[np.object_]],
        rotate_axis: str = 'x'
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """ Converting all attitudes to points on a sphere """
    ary = None
    if len(q_ary.shape) == 1:
        ary = np.empty((3, q_ary.shape[0]))
        for i, q in enumerate(q_ary):
            ary[:,i] = q.to_unit_vector(rotate_axis)[:,0]
    elif len(q_ary.shape) == 2:
        ary = np.empty((3, q_ary.shape[1]))
        for i, q in enumerate(q_ary):
            ary[:,i] = q[0].to_unit_vector(rotate_axis)[:,0]
    else:
        raise ValueError("expected 1- or 2-D array")
    return ary


def scatter(
        q_ary: np.ndarray[Any, np.dtype[np.object_]],
        rotate_axis: str = 'x',
        fig: Optional[mpl.figure.Figure] = None,
        axes: Optional[mpl.axes.Axes] = None,
        scatter_kwargs: Dict[str,Any] = {},
    ) -> mpl.collections.PathCollection:
    """ Plot an array of quaternions as scatter points on a sphere """

    if fig is None:
        fig = plt.figure()
    if axes is None:
        axes = fig.add_subplot(111, projection='3d')

    ary = prepare_quaternion_array_for_plotting(q_ary, rotate_axis = rotate_axis)

    return axes.scatter(ary[0,:], ary[1,:], ary[2,:], **scatter_kwargs)


def plot(
        q_ary: np.ndarray[Any, np.dtype[np.object_]],
        t: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        rotate_axis: str = 'x',
        fig: Optional[mpl.figure.Figure] = None,
        axes: Optional[mpl.axes.Axes] = None,
        plot_kwargs: Dict[str,Any] = {}
    ) -> List[mpl.lines.Line2D]:
    """
    Plot an array of quaternions using lines on the surface of a sphere.

    Optionally, you can provide a t argument to plot as a timeseries,
    where the lowest t is displayed red, and the highest t is
    displayed blue.

    """

    if fig is None:
        fig = plt.figure()
    if axes is None:
        axes = fig.add_subplot(111, projection='3d')

    ary = prepare_quaternion_array_for_plotting(q_ary, rotate_axis=rotate_axis)

    if t is not None:
        tc = (t - t.min()) / (t.max() - t.min())
        c = np.vstack((tc, np.zeros_like(tc), -tc + 1.0))

        lines: List[mpl.lines.Line2D] = []
        for i in range(1,ary.shape[1]):
            lines.extend(axes.plot(ary[0,i-1:i+1], ary[1,i-1:i+1], ary[2,i-1:i+1], c = tuple(c[:,i]), **plot_kwargs))
        return lines
        
    else:
        return axes.plot(ary[0,:], ary[1,:], ary[2,:], **plot_kwargs)

def plot_frame(
        q: Quat,
        r: np.ndarray[Any, np.dtype[np.float64]] = np.zeros((3,1)),
        fig: Optional[mpl.figure.Figure] = None,
        axes: Optional[mpl.axes.Axes] = None,
        axis_size: float = 1.0
    ) -> mpl.axes.Axes:
    """
    Plot a quaternion as a coordinate frame, with red, green, and blue 
    referring to x, y, and z.

    Also accepts an optional position to move the frame to (post-rotation).
    """
    if fig is None:
        fig = plt.figure()
    if axes is None:
        axes = fig.add_subplot(111, projection='3d')

    xh = np.zeros((3,2))
    yh = np.zeros_like(xh)
    zh = np.zeros_like(xh)
    xh[0,1] = axis_size
    yh[1,1] = axis_size
    zh[2,1] = axis_size

    # Transform the frame (rotate it and then move it)
    T = q.to_matrix()
    txh = np.dot(T, xh) + r
    tyh = np.dot(T, yh) + r
    tzh = np.dot(T, zh) + r
    
    axes.plot(txh[0,:], txh[1,:], txh[2,:], c='r')
    axes.plot(tyh[0,:], tyh[1,:], tyh[2,:], c='g')
    axes.plot(tzh[0,:], tzh[1,:], tzh[2,:], c='b')
    
    return axes



"""
This file contains functions for the generation of random
quaternions using various statistical distributions (namely
uniform and gaussian).

It is also useful simply for generating random axes.
"""

from typing import Any, Optional, Callable, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from pyquat._pyquat import Quat
else:
    from _pyquat import Quat

from math import sqrt, cos, sin, pi
import numpy as np

class AxisGeneratorKwargs(TypedDict, total=False):
    max_theta: float
    z_range: float

class AngleGeneratorKwargs(TypedDict, total=False):
    symmetric_range: float

AxisGeneratorType = Callable[..., np.ndarray[Any, np.dtype[np.float64]]]
AngleGeneratorType = Callable[..., float]

def randu(symmetric_range: float = 1.0) -> float:
    """
    Generate a uniform random number with mean 0 and range
    (-symmetric_range, +symmetric_range).  Default argument (1.0)
    gives mean 0 and range (-1, 1).
    """
    
    return float(np.random.rand() * 2.0 - 1.0) * symmetric_range

def uniform_random_axis(
        max_theta: float = 2.0 * pi,
        z_range: float = 1.0
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Generate a unit random axis from a uniform distribution.
    """
    if max_theta == 0.0:
        theta = 0.0
    else:
        theta = float(np.random.rand()) * max_theta

    axis = np.zeros((3,1))
    axis[2] = randu(z_range)
    axis[0] = sqrt(1.0 - axis[2]**2) * cos(theta)
    axis[1] = sqrt(1.0 - axis[2]**2) * sin(theta)
    return axis


def rand(
        axis: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        angle: Optional[float] = None,
        axis_generator: AxisGeneratorType = uniform_random_axis,
        angle_generator: AngleGeneratorType = randu,
        axis_generator_kwargs: AxisGeneratorKwargs = {},
        angle_generator_kwargs: AngleGeneratorKwargs = {},
    ) -> Quat:
    """
    Generate a random quaternion. With all defaults, the quaternion
    will be both random angle and random axis.

    Allows each of axis and angle to be generated or provided; and a
    generator can be provided as well. The default is to use a uniform
    distribution for both axis and angle.

    Any extra arguments to this function are passed on to the axis
    generator.

    """
    if axis is not None and angle is not None:
        raise ValueError("expected non-fixed angle or non-fixed axis or both")
    if axis is None:
        axis  = axis_generator(**axis_generator_kwargs)
    if angle is None:
        angle = angle_generator(**angle_generator_kwargs)
    return Quat.from_angle_axis(angle, *axis)





#!/usr/bin/env python
""" Python C quaternion type for fast attitude and rotation computations.
See:
https://github.com/translunar/pyquat
"""

from setuptools import setup, find_packages
from setuptools.extension import Extension
import os
import numpy

MAJOR = 0
MINOR = 5
TINY  = 1
version='%d.%d.%d' % (MAJOR, MINOR, TINY)

c_quat = Extension('_pyquat',
                   ['pyquat/pyquat.c'],
                   extra_compile_args = ["-std=c99"],
                   include_dirs  = [numpy.get_include()],
                   define_macros = [('MAJOR_VERSION', MAJOR),
                                    ('MINOR_VERSION', MINOR),
                                    ('TINY_VERSION',  TINY)])
setup(name='pyquat',
      version=version,
      description='Python C quaternion type for fast attitude and rotation computations',
      long_description='Python C quaternion type for fast attitude and rotation computations',
      long_description_content_type='text/x-rst',
      author='Juno Woods, Ph.D.',
      author_email='juno@translunar.io',
      url='http://github.com/translunar/pyquat/',
      download_url='https://github.com/translunar/pyquat/archive/v' + version + '.tar.gz',
      include_package_data=True,
      packages=['pyquat', 'pyquat.wahba'],
      ext_modules=[c_quat],
      test_suite='test.pyquat_test_suite',
      license='BSD 3-clause',
      classifiers=[ # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 3 - Alpha',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Topic :: Software Development :: Embedded Systems',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Scientific/Engineering :: Visualization',


          # Pick your license as you wish (should match "license" above)
          'License :: OSI Approved :: BSD License',

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.

          'Programming Language :: C',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: Implementation :: CPython'
      ],
      keywords=['quaternion', 'math', 'maths', 'graphics', 'physics', 'orientation', 'attitude', 'pose', 'geometry', 'visualization', 'visualisation', 'animation', 'game development', 'simulation'],
      install_requires=['numpy', 'scipy']
    )



"""
This module contains quaternion and attitude utility functions
relating to solutions to what is known as Wahba's problem [0],
namely the problem of finding a rotation matrix which brings
one set of vectors into the best possible alignment with a
second set of vectors.

Reference:
[0] Wahba, Grace. 1965. Problems and Solutions: Problem 65--1.
    SIAM Review 7(3): 409.

"""

from typing import Any, Optional, TYPE_CHECKING, TypedDict

import numpy as np
import scipy.linalg as spl

if TYPE_CHECKING:
    from pyquat._pyquat import Quat
else:
    from _pyquat import Quat

class AttitudeProfileKwargs(TypedDict):
    obs: np.ndarray[Any, np.dtype[np.float64]]
    ref: np.ndarray[Any, np.dtype[np.float64]]
    weights: Optional[np.ndarray[Any, np.dtype[np.float64]]]
    q: Optional[Quat]
    cov: Optional[np.ndarray[Any, np.dtype[np.float64]]]
    inverse_cov: Optional[np.ndarray[Any, np.dtype[np.float64]]]

def davenport_matrix(
        B: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        covariance_analysis: bool = False,
        **attitude_profile_kwargs: AttitudeProfileKwargs,
    ) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Compute the Davenport matrix for a given attitude profile.
    Accepts either an attitude profile matrix or the arguments
    for the attitude_profile_matrix() function. Returns a 4x4
    Davenport matrix K.

    If the covariance_analysis argument is True, a slightly modified K
    is returned, which may be used as part of a least squares approach
    (such as the qEKF or the information matrix approach, SOAR). See
    Eqs. 10--11 in [2] for details.

    If you do not provide an attitude profile matrix to this method,
    you must pass it the appropriate arguments for calling
    attitude_profile_matrix().

    References:

    [0] Davenport, P. 1968. A vector approach to the algebra of
        rotations with applications. NASA Technical Report X-546-65-437.

    [1] De Ruiter, A.; Damaren, C.; Forbes, J. 2013. Spacecraft
        Dynamics and Control: An Introduction, 1st Ed. Wiley, West
        Sussex, U.K. pp. 468-471.

    [2] Ainscough, Zanetti, Christian, Spanos. 2015. Q-method extended
        Kalman filter. Journal of Guidance, Control, and Dynamics
        38(4): 752--760.

    Kwargs:
        B                        attitude profile matrix (if not provided, one
                                 will be computed using attitude_profile_kwargs)
        covariance_analysis      defaults to False for the standard Davenport
                                 matrix; if True, returns the Davenport matrix
                                 utilized for least squares
        attitude_profile_kwargs  additional arguments passed to the
                                 attitude_profile_matrix() method if B is not
                                 provided

    Returns:
        A 4x4 matrix.

    """

    if B is None:
        # There are two different versions of the attitude profile matrix
        # computation, one which takes a quaternion and a cov or inverse cov,
        # and another which takes a set of observations and reference vectors.
        # Here we figure out which one to call.
        B = attitude_profile_matrix(**attitude_profile_kwargs)
    
    S  = B + B.T
    mu = np.trace(B)
    zx = B.T - B
    z  = np.array([[zx[2,1]], [zx[0,2]], [zx[1,0]]])

    if covariance_analysis:
        M = S - np.identity(3) * (2*mu)
        upperleft = 0.0
    else:
        M = S - np.identity(3) * mu
        upperleft = mu
    K = np.vstack((np.hstack((np.array([[upperleft]]), z.T)),
                   np.hstack((z,                       M))))
    return K


def quat_attitude_profile_matrix(
        q: Quat,
        cov: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        inverse_cov: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
    ) -> np.ndarray[Any, np.dtype[np.float64]]:

    """
    FIXME: Needs better documentation.

    See reference given in attitude_profile_matrix() for documentation.
    """
    if inverse_cov is None:
        inverse_cov = spl.inv(cov)

    tr = np.trace(inverse_cov)
    return np.dot(np.identity(3) * tr * 0.5 - inverse_cov, q.to_matrix())

def attitude_profile_matrix(
        obs: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        ref: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        weights: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        q: Optional[Quat] = None,
        cov: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
        inverse_cov: Optional[np.ndarray[Any, np.dtype[np.float64]]] = None,
    ) -> np.ndarray[Any, np.dtype[np.float64]]:

    """"There exists a unique orthogonal matrix A which satisfies
   
       A \hat{r}_i = \hat{s}_i for i in (1,2,3)

    which is given by

       A = \sum_{i=1}^{3} \hat{s}_i \hat{r}_i^\top
    
    " [0] and is called the attitude profile matrix. Note that
    observation vector \hat{s}_i is 3x1 and reference vector
    \hat{r}_i^\top is 1x3 in the above. Also permitted is a weight on
    each vector observation describing the relative precision of the
    sensor which provided it; see Eq. 5 from [1] and Eq. 10 from [2].

    Compute the attitude profile matrix B using either vector
    observations and reference vectors or a quaternion and 3x3
    attitude error covariance.

    weights should be positive and sum to 1.0.

    References:

    [0] Shuster, Oh. 1981. Three-axis attitude determination from
        vector observations. Journal of Guidance and Control 4(1):
        70--77.

    [1] Mortari. 1997. ESOQ-2 single-point algorithm for fast optimal
        spacecraft attitude determination. In Advances in the
        Astronautical Sciences.

    [2] Ainscough, Zanetti, Christian, Spanos. 2015. Q-Method extended
        Kalman filter. Journal of Guidance, Control, and Dynamics
        38(4): 752--760.

    Kwargs:
        q    

    """

    if q is not None:
        return quat_attitude_profile_matrix(q=q, cov=cov, inverse_cov=inverse_cov)

    # Make sure we have a weights vector since this is a weighted least
    # squares problem.
    n = obs.shape[1]
    if weights is None:
        weights = np.ones(n) / float(n)

    B = np.zeros((3,3))
    for ii in range(0, n):
        B += np.dot(obs[0:3,ii].reshape((3,1)), ref[0:3,ii].reshape((1,3))) * weights[ii]
        
    return B



import typing
from typing import TYPE_CHECKING

import numpy as np

from test.assertions import QuaternionTest

if TYPE_CHECKING:
    import pyquat.pyquat as pq
    import pyquat.pyquat.wahba.esoq as esoq
    import pyquat.pyquat.wahba as wahba
else:
    import pyquat as pq
    import pyquat.wahba.esoq as esoq
    import pyquat.wahba as wahba
        
class TestWahbaESOQ(QuaternionTest):
    def test_attitude_profile_matrix_from_quaternion(self) -> None:
        """attitude_profile_matrix() doesn't raise errors when given a quaternion measurement and a covariance"""
        q   = pq.identity()
        cov = np.identity(3)
        B   = wahba.quat_attitude_profile_matrix(q, cov)
        # Needs actual test here
        
    
    def test_davenport_matrix_from_quaternion(self) -> None:
        """The Davenport matrix produced by a quaternion is the same as that produced by attitude_profile_matrix() called on that same quaternion"""
        q   = pq.identity()
        cov = np.identity(3)
        K1  = wahba.davenport_matrix(q = q, cov = cov)

        B   = wahba.quat_attitude_profile_matrix(q, cov)
        K2  = wahba.davenport_matrix(B)
        np.testing.assert_array_equal(K1, K2)

    
    def test_davenport_eigenvalues(self) -> None:
        """davenport_eigenvalues() produces eigenvalues in the appropriate range"""
        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[0.0, 0.0],
                        [1.0, 0.0],
                        [0.0, 1.0]])
        B = wahba.attitude_profile_matrix(obs = obs, ref = ref)
        irot = esoq.sequential_rotation_helper(B)
        K = wahba.davenport_matrix(B)
        l = esoq.davenport_eigenvalues(K, B, n_obs = 2)
        self.assertLessEqual(-1.0 - 1e-6, l[3])
        self.assertLessEqual(l[3], l[2])
        self.assertLessEqual(l[2], l[1])
        self.assertLessEqual(l[1], l[0])
        self.assertLessEqual(l[0], 1.0 + 1e-6)

    def test_trace_adj(self) -> None:
        """The adjugate trace is as expected"""
        K = np.array([[-1.0, 0, 2, -2],
                      [0, 3, 0, 0],
                      [2, 0, -1, -2],
                      [-2, 0, -2, -1]])
        ta1 = esoq.trace_adj(K)
        ta2 = np.trace(K)
        self.assertEqual(ta1, ta2)

    def test_esoq2_0_rotation(self) -> None:
        """ESOQ2 works properly with no rotation between ref and obs"""
        #Test borrowed from https://github.com/muzhig/ESOQ2/blob/master/test.cpp
        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        self.assert_esoq2_two_observations_correct(ref = ref, obs = obs, decimal=12)

    def test_esoq2_90_z_rotation(self) -> None:
        """ESOQ2 works properly with a 90-degree about-z rotation between ref and obs"""

        # Test borrowed from https://github.com/muzhig/ESOQ2/blob/master/test.cpp

        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[0.0, 0.0],
                        [1.0, 0.0],
                        [0.0, 1.0]])
        self.assert_esoq2_two_observations_correct(ref = ref, obs = obs, decimal=12)

    def test_esoq2_90_x_rotation(self) -> None:
        """ESOQ2 works properly with a 90-degree about-x rotation between ref and obs"""
        
        # Test borrowed from https://github.com/muzhig/ESOQ2/blob/master/test.cpp

        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[1.0, 0.0],
                        [0.0, -1.0],
                        [0.0, 0.0]])
        self.assert_esoq2_two_observations_correct(ref = ref, obs = obs, decimal=12)

    def test_esoq2_90_y_rotation(self) -> None:
        """ESOQ2 works properly with a 90-degree about-y rotation between ref and obs"""
        
        # Test borrowed from https://github.com/muzhig/ESOQ2/blob/master/test.cpp
        ref = np.array([[1.0, 0.0],
                        [0.0, 0.0],
                        [0.0, 1.0]])
        obs = np.array([[0.0, 1.0],
                        [0.0, 0.0],
                        [-1.0, 0.0]])
        self.assert_esoq2_two_observations_correct(ref = ref, obs = obs, decimal=12)

        
    def test_esoq2_actual_rotation(self) -> None:
        """Verifies that ESOQ2 produces the expected result for an arbitrary set"""
        ref_d = np.array([[-0.13745816],
                          [ 0.44258304],
                          [ 0.88612951]])
        ref_e = np.array([[-0.27904739],
                          [ 0.8410828 ],
                          [-0.46337056]])
        obs_d = np.array([[-0.65361096],
                          [ 0.38250994],
                          [ 0.65305348]])
        obs_e = np.array([[ 0.37343209],
                          [ 0.91352261],
                          [-0.16132242]])

        ref   = np.hstack((ref_d, ref_e))
        obs   = np.hstack((obs_d, obs_e))
        self.assert_esoq2_two_observations_correct(ref = ref, obs = obs, decimal=8)        




import typing
from typing import TYPE_CHECKING

import numpy as np
from scipy import linalg
from test.assertions import QuaternionTest

if TYPE_CHECKING:
    import pyquat.pyquat as pq
else:
    import pyquat as pq


PLOT = False

class TestPyquat(QuaternionTest):
    """Tests basic functionality on pyquat and the pyquat Quaternion type 'Quat' written in C"""

    def test_mean(self) -> None:
        """The mean quaternion is computed properly"""
        self.assert_equal(
            pq.mean(np.array([[1.0, 0.0, 0.0, 0.0],
                              [1.0, 0.0, 0.0, 0.0]]).T), 
            pq.Quat(1.0, 0.0, 0.0, 0.0))
        
        self.assert_almost_equal_components(
            pq.mean(np.array([[1.0, 0.0, 0.0, 0.0],
                              [0.0, 1.0, 0.0, 0.0]]).T),
            pq.Quat(0.70710678118654757, 0.70710678118654757, 0.0, 0.0),
            delta=1e-12)
        
        self.assert_almost_equal_components(
            pq.mean(np.array([pq.Quat(1.0, 0.0, 0.0, 0.0),
                              pq.Quat(0.0, 1.0, 0.0, 0.0)])),
            pq.Quat(0.70710678118654757, 0.70710678118654757, 0.0, 0.0),
            delta=1e-12)

    def test_positive_scalar(self) -> None:
        """Quaternions initialize with positive scalar component regardless of input"""
        self.assert_equal(pq.Quat(-1.0, 0.0, 0.0, 0.0), pq.identity())

        self.assert_equal(pq.Quat(-1.0, 2.0, 3.0, 4.0).normalized(),
                          pq.Quat(1.0, -2.0, -3.0, -4.0).normalized())

    def test_identity(self) -> None:
        """The identity quaternion is properly constructed"""
        self.assert_equal_as_matrix(
            pq.identity(),
            np.identity(3))
        self.assert_equal_as_quat(
            pq.identity(),
            np.identity(3))

    def test_symmetric_matrix_conversion(self) -> None:
        """Conversion to and from a matrix is symmetric"""
        q = pq.Quat(0.4, -0.3, 0.2, -0.1)
        q.normalize()
        self.assert_almost_equal_as_quat(
            q,
            q.to_matrix())
        self.assert_almost_equal_as_quat(
            q.conjugated(),
            q.to_matrix().T)

    def test_symmetric_rotation_vector_conversion(self) -> None:
        """Conversion to and from a rotation vector is symmetric"""
        q = pq.Quat(0.4, -0.3, 0.2, -0.1)
        q.normalize()
        self.assert_almost_equal_components(q, pq.Quat.from_rotation_vector(q.to_rotation_vector()))
        

    def test_symmetric_angle_axis_conversion(self) -> None:
        """Conversion to and from an angle-axis is symmetric"""
        q = pq.Quat(0.4, -0.3, 0.2, -0.1)
        q.normalize()
        phi = q.to_rotation_vector()
        angle = linalg.norm(phi)
        phi_hat = phi / angle
        self.assert_almost_equal_components(q, pq.Quat.from_angle_axis(angle, phi_hat[0], phi_hat[1], phi_hat[2]))

    def test_angle_axis_with_theta(self) -> None:
        """Accepts the alternate parameters for from_angle_axis"""
        q = pq.Quat(0.4, -0.3, 0.2, -0.1)
        q.normalize()
        phi = q.to_rotation_vector()
        angle = linalg.norm(phi)
        phi_hat = phi / angle
        x, y, z = phi_hat[0], phi_hat[1], phi_hat[2]
        # Now find theta using x, y, and z
        # x2 = (1 - z2) * cos2(theta)
        # y2 = (1 - x2) * sin2(theta)
        theta = np.arctan2(y, x)
        self.assert_almost_equal_components(q, pq.Quat.from_angle_axis(angle, z=z, theta=theta))

    def test_symmetric_conjugate(self) -> None:
        """Quaternion conjugation is symmetric"""
        q = pq.Quat(0.4, -0.3, 0.2, -0.1)
        qT = q.conjugated()
        self.assert_equal(q, qT.conjugated())

    def test_small_rotation_vector(self) -> None:
        """Construction directly from a small rotation vector produces the same quaternion as conversion of a rotation vector to a matrix and then to a quaternion"""
        v = np.zeros((3,1)) * 3.0 / np.sqrt(3.0)
        q = pq.Quat.from_rotation_vector(v)
        self.assert_equal(q, pq.identity())
        T = pq.rotation_vector_to_matrix(v)
        np.testing.assert_array_equal(T, q.to_matrix())

    def test_large_rotation_vector(self) -> None:
        """Construction directly from a large rotation vector produces an almost-identical quaternion as conversion of a rotation vector to a matrix and then to a quaternion
        """        
        v = np.array([[3.0, 2.0, 1.0]]).T
        T1 = pq.Quat.from_rotation_vector(v).to_matrix()
        T2 = pq.rotation_vector_to_matrix(v)
        np.testing.assert_array_almost_equal(T1, T2)

    def test_multiplication(self) -> None:
        """Quaternion multiplication produces the same result as attitude matrix multiplication"""
        qAB = pq.Quat(0.4, -0.3, 0.2, -0.1)
        qAB.normalize()
        qBC = pq.Quat(0.2, 0.3, -0.4, 0.5)
        qBC.normalize()
        qAC = qBC * qAB
        qAC.normalize()

        tAB = qAB.to_matrix()
        tBC = qBC.to_matrix()
        tAC = np.dot(tBC, tAB)
        self.assert_almost_equal_as_matrix(qAC, tAC)

    def test_normalize(self) -> None:
        """Normalization works properly for typical values """
        q0 = pq.Quat(4.0, 3.0, 2.0, 1.0)
        q1 = pq.Quat(4.0, 3.0, 2.0, 1.0)
        q5 = pq.Quat(4.0, 3.0, 2.0, 1.0)
        q2 = q1.normalized()

        # Test that normalized() changed q2 and not q1
        self.assert_not_equal(q1, q2)
        self.assert_equal(q0, q1)

        # Test that normalized_large() works
        q4 = q1.normalized_large()
        self.assert_equal(q2, q4)

        # Now test that normalize() changes q1
        q1.normalize()
        self.assert_not_equal(q0, q1)
        self.assert_equal(q1, q2)

        # Test that normalize_large() changes q4 correctly
        q5.normalize_large()
        self.assert_equal(q1, q5)

        # Now test that normalize does what we expect it to do.
        v = np.array([[4.0, 3.0, 2.0, 1.0]]).T
        v /= linalg.norm(v)
        q3 = pq.Quat(v[0], v[1], v[2], v[3])
        self.assert_equal(q1, q3)

        # Now test that normalize handles invalid quaternions.
        q3 = pq.Quat(0.0, 0.0, 0.0, 0.0)
        self.assert_equal(q3.normalized(), pq.identity()) # out-of-place test
        self.assert_equal(q3, pq.Quat(0.0, 0.0, 0.0, 0.0))
        q3.normalize()
        self.assert_equal(q3, pq.identity()) # in-place test

    def test_normalize_large(self) -> None:
        """Overflow is avoided in normalization"""
        q_max = 3.9545290113758423e+256
        q0 = pq.Quat(-9.6241008572232875e+255,
                   q_max,
                  -2.3364730154155227e+255,
                  -2.2751942430616868e+256)
        v0 = q0.to_vector()
        v1 = v0 / q_max
        v1_mag = np.sqrt(v1[0,0]**2 + v1[1,0]**2 + v1[2,0]**2 + v1[3,0]**2)
        v2 = v1 / v1_mag
        q1 = pq.Quat(v2[0], v2[1], v2[2], v2[3]) 
        q2 = q0.normalized_large()
        self.assert_equal(q1, q2)

    def test_conjugate(self) -> None:
        """In-place and out-of-place conjugation works as expected"""
        q0 = pq.Quat(4.0, -3.0, -2.0, -1.0)
        q1 = pq.Quat(4.0,  3.0,  2.0,  1.0)

        # Test out-of-place
        self.assert_equal(q0, q1.conjugated())
        self.assert_not_equal(q0, q1)

        # Test in-place
        q1.conjugate()
        self.assert_equal(q0, q1)

    def test_skew(self) -> None:
        """The skew matrix is produced from a 3x1 vector correctly in C code"""
        w = np.array([[0.03, 0.02, 0.01]]).T

        # old method:
        skv = np.roll(np.roll(np.diag(w.flatten()), 1, 1), -1, 0)
        wx  = skv - skv.T
        
        np.testing.assert_array_equal(wx, pq.skew(w))        

    def test_propagate(self) -> None:
        """Simple propagation with small angular velocities produces the same result for matrices and quaternions"""
        dt = 0.01
        q0 = pq.Quat(1.0, 0.0, 0.0, 0.0)
        w0 = np.array([[3.0, 2.0, 1.0]]).T
        q1 = pq.propagate(q0, w0, dt)
        phi1 = q1.to_rotation_vector()
        phi2 = w0 * dt
        q2   = pq.from_rotation_vector(phi2)
        self.assert_almost_equal(q1, q2, decimal=6)

        # Test quaternion propagation
        q3 = pq.propagate(q0, w0, dt)
        self.assert_equal(q1, q3)

        # Compare the directed cosine matrix result
        T0   = q0.to_matrix()
        w0x  = pq.skew(w0)
        Tdot = np.dot(T0, w0x)
        T1   = np.identity(3) - Tdot * dt # dT = (I - [phi x])
        qT1  = pq.from_matrix(T1)
        self.assert_almost_equal(q1, qT1)

        # Compare to the additive result
        q4 = pq.propagate_additively(q0, w0, dt)
        self.assert_almost_equal(q4, qT1)

        # Compare to the matrix propagation
        # Note: Matrix propagation is new and I'm not totally sure it's
        # correct yet, so this just verifies that it looks reasonable.
        # If it's correct, Tr4 should be the most accurate.
        T0   = q0.to_matrix()
        Tr1  = pq.matrix_propagate(T0, w0, dt, r=1)
        Tr2  = pq.matrix_propagate(T0, w0, dt, r=2)
        Tr3  = pq.matrix_propagate(T0, w0, dt, r=3)
        Tr4  = pq.matrix_propagate(T0, w0, dt, r=4)
        np.testing.assert_equal(T1, Tr1)
        np.testing.assert_almost_equal(Tr1, Tr4, decimal=3)
        np.testing.assert_almost_equal(Tr2, Tr4, decimal=5)
        np.testing.assert_almost_equal(Tr3, Tr4, decimal=7)

        # Test that zero-propagation is identical no matter the degree.
        T0r4 = pq.matrix_propagate(T0, np.zeros(3), dt, r=4)
        np.testing.assert_equal(T0, T0r4)

    def test_0_propagate(self) -> None:
        """Quaternion propagation with a 0 angular velocity does not lead to a zero division
        """
        dt = 0.01
        q0 = pq.Quat(1.0, 0, 0, 0)
        w0 = np.zeros((3,1))
        q1 = pq.propagate(q0, w0, dt)

        self.assert_equal(q0, q1)

    def test_rk4_integration(self) -> None:
        """Runge-Kutta 4 integration does not produce errors"""
        dt = 0.05
        q = pq.Quat(1.0, 2.0, 3.0, 4.0).normalized()
        w = np.array([[0.03, 0.02, 0.01]]).T
        J = np.diag([200.0, 200.0, 100.0])
        J_inv = linalg.inv(J)

        # Test propagation using RK4
        q1, w1 = pq.step_rk4(q,  w,  dt,     J = J, J_inv = J_inv)
        qa, wa = pq.step_rk4(q,  w,  dt*0.5, J = J, J_inv = J_inv)
        q2, w2 = pq.step_rk4(qa, wa, dt*0.5, J = J, J_inv = J_inv)

    def test_expm(self) -> None:
        """The closed form expression for expm from a quaternion and angular velocity produces the same result as linalg.expm()"""
        B1    = 13/51.0
        dt    = 0.05
        w     = np.array([[0.03, 0.02, 0.01]]).T
        J     = np.diag([200.0, 200.0, 100.0])
        J_inv = linalg.inv(J)
        wk1   = pq.wdot(w, J)
        qk1   = pq.state_transition_matrix(w)
        expm1 = pq.expm(w, dt * B1)
        expm2 = linalg.expm(qk1 * (B1 * dt))
        np.testing.assert_array_almost_equal(expm1, expm2)
        
    def test_cg_integration(self) -> None:
        """CG3, CG4, and RK4 integration produce reasonably similar results"""
        dt = 0.1
        q = pq.Quat(1.0, 2.0, 3.0, 4.0).normalized()
        w = np.array([[0.3, 0.2, 0.1]]).T
        J = np.diag([200.0, 200.0, 100.0])

        # Test propagation using RK4
        q1, w1 = pq.step_rk4(q, w, dt, J = J, w_dynamics = pq.wdot)

        # Test propagation using CG3
        q2, w2 = pq.step_cg3(q, w, dt, J = J, w_dynamics = pq.wdot)

        # Test propagation using CG4
        q3, w3 = pq.step_cg4(q, w, dt, J = J, w_dynamics = pq.wdot)

        self.assert_almost_equal(q2, q3)
        self.assert_almost_equal(q1, q2)
        np.testing.assert_array_almost_equal(w2, w3)
        np.testing.assert_array_almost_equal(w1, w2)

        # Test propagation without angular velocity dynamics using each method
        q4, w4 = pq.step_rk4(q, w, dt, w_dynamics = None)
        q5, w5 = pq.step_cg3(q, w, dt, w_dynamics = None)
        q6, w6 = pq.step_cg4(q, w, dt, w_dynamics = None)

        self.assert_almost_equal(q4, q5)
        self.assert_almost_equal(q5, q6)
        np.testing.assert_array_equal(w, w4)
        np.testing.assert_array_equal(w, w5)
        np.testing.assert_array_equal(w, w6)

    def test_cg_integration_tiny_steps(self) -> None:
        """CG4 integration produces reasonably similar results over many small steps as it would over one big step"""

        # This test mostly exists to show that integration is not very accurate.
        
        q0 = pq.Quat(1.0, 2.0, 3.0, 4.0).normalized()
        
        w = np.array([[-1.0, -3.0, -4.0]]).T * (np.pi / 180.0)
        
        dt_small = 0.001
        dt_large = 0.1
        t        = 1.0

        dt_total = 0.0
        q_large = q0.copy()
        while dt_total < t:
            q_large, w_out = pq.step_cg4(q_large, w, dt_large, w_dynamics = None)
            dt_total += dt_large
        q_large.normalize()
            
        dt_total = 0.0
        q_small  = q0.copy()
        while dt_total < t:
            q_small, w_out = pq.step_cg4(q_small, w, dt_small, w_dynamics = None)
            dt_total += dt_small
        q_small.normalize()

        self.assert_not_almost_equal(q_large, q0)
        self.assert_not_almost_equal(q_small, q0)

        self.assert_almost_equal(q_small, q_large, decimal=2) # FAIRLY INACCURATE!

    def test_integration_handles_zero(self) -> None:
        """CG3, CG4, and RK4 integration correctly hadnle an angular velocity of  0"""
        dt = 0.1
        q = pq.identity()
        w = np.zeros((3,1))
        J = np.diag([200.0, 200.0, 100.0])
        q1, w1 = pq.step_rk4(q, w, dt, J = J)
        q2, w2 = pq.step_cg3(q, w, dt, J = J)
        q3, w3 = pq.step_cg4(q, w, dt, J = J)

        self.assert_equal(q, q1)
        self.assert_equal(q, q2)
        self.assert_equal(q, q3)

    def test_big_omega(self) -> None:
        """The C big_omega() method produces the expected result"""
        w = np.array([[0.03, 0.02, 0.01]]).T
        W = pq.big_omega(w)

        W1 = np.array([[0.0,    -w[0,0], -w[1,0], -w[2,0]],
                       [w[0,0],  0.0,     w[2,0], -w[1,0]],
                       [w[1,0], -w[2,0],  0.0,     w[0,0]],
                       [w[2,0],  w[1,0], -w[0,0],  0.0]])
        
        np.testing.assert_array_equal(W, W1)

    def test_tobytes_and_fromstring(self) -> None:
        """Conversion to and from strings is symmetric"""
        q1 = pq.Quat(1.0, 2.0, 3.0, 4.0).normalized()
        s  = q1.tobytes()
        q2 = pq.fromstring(s, dtype=np.float64, count=4)
        self.assert_equal(q1, q2)

    def test_fromstring_with_sep(self) -> None:
        """Conversion from a string with separator character works properly"""
        q1 = pq.fromstring("1.0 2.0 3.0 4.0", sep=' ').normalized()
        q2 = pq.Quat(1.0, 2.0, 3.0, 4.0).normalized()
        self.assert_equal(q1, q2)

    def test_copy(self) -> None:
        """Copies the quaternion"""
        q1 = pq.Quat(1.0, 2.0, 3.0, 4.0).normalized()
        q2 = q1.copy()
        q2.w = 0.0
        q2.normalize()
        self.assert_not_equal(q1, q2)
        self.assert_equal(q1, pq.Quat(1.0, 2.0, 3.0, 4.0).normalized())

    def test_dot(self) -> None:
        """Dot product of two quaternions should be the dot product of their
        respective to_vector() methods"""
        q1 = pq.Quat(1.0, 0.0, 0.0, 0.0)
        q2 = pq.Quat(1.0, 1.0, 0.0, 0.0).normalized()
        self.assertEqual(q1.dot(q2),
                         q1.to_vector().reshape(4).dot(q2.to_vector().reshape(4)))
        
    def test_lerp(self) -> None:
        """Linear interpolation works as expected for a 45-degree angle"""
        q1 = pq.Quat(1.0, 0.0, 0.0, 0.0)
        q2 = pq.Quat(1.0, 1.0, 0.0, 0.0).normalized()

        # Check half-way point equidistant from inputs
        q3 = q1.lerp(q2, 0.5)
        self.assertAlmostEqual(q3.dot(q1), q3.dot(q2))

        # Check that each input is 22.5 degrees from the output
        q3n = q3.normalized()
        self.assertAlmostEqual(np.arccos(q3n.dot(q2)) * 180.0 / np.pi, 22.5)
        self.assertAlmostEqual(np.arccos(q3n.dot(q1)) * 180.0 / np.pi, 22.5)

        # Check that q1 lerped with itself gives itself.
        q4 = q1.lerp(q1.copy(), 1.0)
        self.assert_equal(q1, q4)
        q4 = q1.lerp(q1.copy(), 0.0)
        self.assert_equal(q1, q4)
        q4 = q1.lerp(q1.copy(), 0.5)
        self.assert_equal(q1, q4)

        # Check end-ranges of the t parameter
        q5 = q1.lerp(q2, 0.0)
        self.assert_equal(q5, q1)

        q6 = q1.lerp(q2, 1.0)
        self.assert_equal(q6, q2)

    def test_slerp(self) -> None:
        """Spherical linear interpolation works as expected for a 45-degree angle"""
        q1 = pq.Quat(1.0, 0.0, 0.0, 0.0)
        q2 = pq.Quat(1.0, 1.0, 0.0, 0.0).normalized()

        # Check half-way point equidistant from inputs
        q3 = q1.slerp(q2, 0.5)
        self.assertAlmostEqual(q3.dot(q1), q3.dot(q2))

        # Check that each input is 22.5 degrees from the output
        self.assertAlmostEqual(np.arccos(q3.dot(q2)) * 180.0 / np.pi, 22.5)
        self.assertAlmostEqual(np.arccos(q3.dot(q1)) * 180.0 / np.pi, 22.5)

        # Check that q1 lerped with itself gives itself.
        q4 = q1.slerp(q1.copy(), 1.0)
        self.assert_equal(q1, q4)
        q4 = q1.slerp(q1.copy(), 0.0)
        self.assert_equal(q1, q4)
        q4 = q1.slerp(q1.copy(), 0.5)
        self.assert_equal(q1, q4)

        # Check end-ranges of the t parameter
        q5 = q1.slerp(q2, 0.0)
        self.assert_equal(q5, q1)

        q6 = q1.slerp(q2, 1.0)
        self.assert_equal(q6, q2)

    def test_slerp_with_lerp(self) -> None:
        """slerp should resort to lerp when a lerp_threshold is set"""
        q1  = pq.Quat(1.0, 0.0, 0.0, 0.0)
        q2  = pq.Quat(1.0, 0.1, 0.0, 0.0).normalized()
        dot = q1.dot(q2)

        lerp  = q1.lerp(q2, 0.5).normalized()
        slerp = q1.slerp(q2, 0.5)
        mixed = q1.slerp(q2, 0.5, lerp_threshold = 0.9)
        self.assert_equal(lerp, mixed)
        self.assert_not_equal(lerp, slerp)


    def test_rotate(self) -> None:
        """Rotation of a vector using a quaternion is equivalent to standard matrix-vector multiplication over many random test quaternions and vectors (to 14 decimal places)"""
        import pyquat.random as pqr
        
        for ii in range(0, 100):
            q = pqr.rand()
            T = q.to_matrix()
            v = pqr.uniform_random_axis() * 10.0 # vector of length 10
            qvq = q.rotate(v)
            Tv  = T.dot(v)
            np.testing.assert_almost_equal(Tv, qvq, decimal=14)


    def test_small_propagation(self) -> None:
        """Propagation of non-identity quaternions by small amounts produces same result as 4th-order matrix propagation."""
        q0 = pq.Quat(0.97020921924687786, 0.00092059174936653205, 0.0073268579434615034, 0.24215602522314206)
        w = np.array([[-1.23692754e-06],
                      [  4.35427879e-23],
                      [ -7.72662863e-07]])
        dt = 0.01
        T0 = q0.to_matrix()
        q1 = pq.propagate(q0, w, dt)
        T1 = pq.matrix_propagate(T0, w, dt, r=4)
        qT1 = pq.from_matrix(T1)
        np.testing.assert_almost_equal(qT1.to_vector(), q1.to_vector(), decimal=16)

        # Note that we cannot compare the results as quaternions, as there is too
        # much error in the multiplication operation. We must use
        #
        #   dT = T0.T.dot(T1)
        #
        # rather than
        #
        #   dq = q1 * q0.conjugated()
        #


    def test_interp(self) -> None:
        """Numpy-style interpolation"""

        # First set up a series
        xp = [-0.1, 0.0, 0.1]
        w0 = np.array([1.0, 2.0, 3.0])
        w1 = np.array([0.9, 1.9, 3.1])
        q0 = pq.identity()
        q1 = pq.propagate(q0, w0, 0.1)
        q2 = pq.propagate(q1, w1, 0.1)
        fp = [q0, q1, q2]

        # Now interpolate it
        x = [-0.15, -0.105, -0.1, -0.09, -0.05, 0.05, 0.1, 0.15]
        f = pq.interp(x, xp, fp)
        self.assert_equal(f[2], q0)
        self.assert_equal(f[6], q2)

        # Test by inspection:
        # FIXME: This needs to be more rigorously tested.
        if PLOT == True:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.add_subplot(111)
            q0v = q0.to_vector()
            q1v = q1.to_vector()
            q2v = q2.to_vector()
            ax.plot(xp, [q0v[0], q1v[0], q2v[0]])
            ax.plot(xp, [q0v[1], q1v[1], q2v[1]])
            ax.plot(xp, [q0v[2], q1v[2], q2v[2]])
            ax.plot(xp, [q0v[3], q1v[3], q2v[3]])

            for ii in range(0, len(x)):
                ax.scatter([x[ii], x[ii], x[ii], x[ii]], f[ii].to_vector(), s=2)
            plt.show() # type: ignore

    def test_Xi(self) -> None:
        """Xi matrix construction"""
        q = pq.Quat(1.0, 2.0, -3.0, 4.0).normalized()
        Xi = q.big_xi()
        qv = np.array([q.x, q.y, q.z])
        np.testing.assert_equal(Xi[0,0:3], -qv)
        np.testing.assert_equal(Xi[1:4,0:3], np.identity(3) * q.w + pq.skew(qv))
